# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from typing import Optional

import pandas as pd
import plotly.express as px


def gen_pcnt_barplot(df: pd.DataFrame, x_col: str, title: Optional[str] = None):
    """
    Generates a percentage bar plot for distinct values in a column.

    Args:
        df: The pandas DataFrame containing the data.
        x_col: The column to use for the x-axis (categorical).
        title: Optional title of the plot. If None, a default title is used.

    Returns:
        A plotly.graph_objects.Figure object.
    """

    # Calculate value counts and percentages
    value_counts = df[x_col].value_counts(normalize=True) * 100
    bar_df = pd.DataFrame(
        {x_col: value_counts.index.astype(str), "percentage": value_counts.values}
    )

    # Determine title
    if title is None:
        title = f"Percentage Distribution of {x_col}"

    # Create the bar plot
    fig = px.bar(
        bar_df,
        x=x_col,
        y="percentage",
        title=title,
        labels={"percentage": "Percentage"},
    )

    return fig



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""This module contains generic constants for models and algos."""

VARIANT_COL = "variant"
"""
Denotes a column in data which includes experiment variant id.
This could be a tuple for multi-experiments.
The expected type is either `str` or `tuple[str]`.
"""

VARIANT_COUNT_COL = "variant_count"
"""
The count of units for that given variants.
The expected type is `int`.
"""

VARIANT_PERCENT_COL = "variant_percent"
"""
The percent of units for that given variants.
The expected type is `float`.
"""

TRIGGER_STATE_COL = "trigger_state"
"""Denotes a column in data which includes the status of a unit of experiment in terms of triggering.
This is either a `bool` for univariate experiments or `tuple[bool]` for multi-experiments.
"""

TRIGGER_STATE_COUNT_COL = "trigger_state_count"
"""The count of units for that given trigger state.
The expcted type is `int`.
"""

TRIGGER_STATE_PERCENT_COL = "trigger_state_percent"
"""The percent of units for that given trigger state.
The expected type is `float`.
"""

TRIGGER_STATE_OVERALL_COL = "trigger_state_overall"
"""Denotes a column in data which includes the overall trigger state of a unit of experiment.
This is a `bool` column.
"""

TRIGGER_STATE_OVERALL_PERCENT_COL = "trigger_state_overall_weight"
"""Denotes a column in data which includes the overall trigger state of a unit of experiment as a percentage.
This is a `float` column.
"""

TRIGGER_STATE_OVERALL_COUNT_COL = "trigger_state_overall_count"
"""Denotes a column in data which includes the overall trigger state of a unit of experiment as a count.
This is an `int` column.
"""

VARIANT_OVER_TRIGGERED_PERCENT_COL = "variant_over_triggered_pcnt"
"""
Denotes a column in data which includes the percent of the corresponding triggered for the given variant.
This will constitue a distribution of variant occurrence conditional on the given trigger stats.
For example see the following two examples:

    - the trigger state is (True, False).
        Then the corresponding variants could be these three:
        ("v1", "nan") and ("v2", "nan") and ("control", "nan")
        each covering 33.33% of the triggered (sums up to 100.00%).
    - the trigger state is (True, True).
        Then correponding variants could be these six:
        ("v1", "enabled"), ("v2", "enabled"), ("control", "enabled"),
        ("v1", "control"), ("v2", "control"), ("control", "control")
        each covering variuos percentages which also sums up to 100.00%

This is useful information to understand the experiment assignments and can be
used in calculating the impact of the experiment across population (site-wide impact) after experiment is launched.
Note that such information could be available during the experiment design as the assignments are decided.
The two distributions will not be exactly the same due to randomness in the assignment procedure.
"""

TIME_COL = "datepartition"
"""
Denotes a column which includes timestamp / date information.
"""

TRIGGER_TIME_COL = "trigger_date"
"""
Experiment trigger date column.
We assume this column is a tuple column.
"""

MIN_TRIGGER_TIME_COL = "min_trigger_date"
"""
This is the minimum trigger date across experiments.
This is an added column to `expt_df`.
"""

MEAN_COL = "mean"
"""
Denotes a column in data which includes the mean of a metric.
"""

SD_COL = "sd"
"""Denotes a column in data which includes the standard deviation of a metric."""

VAR_COL = "var"
"""Denotes a column in data which includes the variance of a metric."""

SAMPLE_COUNT_COL = "sample_count"
"""
This often maps to `VARIANT_COUNT_COL` in the context of experimentation.
Denotes a column in data which includes the count of units (typically experiment units).
The sample here referes to observed units rather than a larger pool which can be triggered units.

The number of triggerd units (see `TRIGGERED_COUNT_COL`) can be higher than the sample count,
because that might include other variants / conterfactuals in the experiment.

For example for an experiment with arms "control", "v1", "v2".

    - the count for "v1" can be 30
    - the triggered count can be 100 because it includes the observed counts for "v2" and "control" as well.

Also note that situation might be more complicated for multi-experiments but the same logic applies.

For example assume

Expt 1 has three variants v1, v2, control
Expt 2 has three variants enabled, control

a launch (v1, enabled): impact will be in three partitions

    - (v1, nan), (nan, enabled), (v1, enabled)

then these need to compared with (respective to their order)

    - (control, nan), (nan, control), (control, control)

For example for (v1, nan) v.s. (control, nan) each of these have an observed count
and triggered count.

The triggered count for (v1, nan) will consist of the sum of the observed counts for

    - (control, nan)
    - (v1, nan)
    - (v2, nan)

"""

TRIGGERD_COUNT_COL = "triggered_count_col"
"""
This often maps to `TRIGGER_STATE_COUNT_COL` in the context of experimentation.
Denotes a column in data which includes the count of units for all the possible variants
which are to choose from in an experiment.
The number of triggerd units can be higher than the sample count (see `SAMPLE_COUNT_COL`),
because that might include other variants / conterfactuals  in the experiment.

For example for an experiment with arms "control", "v1", "v2".

    - the count for "v1" can be 30
    - the triggered count can be 100 because it includes the observed counts for "v2" and "control" as well.

Also note that situation might be more complicated for multi-experiments but the same logic applies.

For example assume

Expt 1 has three variants v1, v2, control
Expt 2 has three variants enabled, control

a launch (v1, enabled): impact will be in three partitions

    - (v1, nan), (nan, enabled), (v1, enabled)

then these need to compared with (respective to their order)

    - (control, nan), (nan, control), (control, control)

For example for (v1, nan) v.s. (control, nan) each of these have an observed count
and triggered count.

The triggered count for (v1, nan) will consist of the sum of the observed counts for

    - (control, nan)
    - (v1, nan)
    - (v2, nan)
"""

SUM_COL = "sum"
"""Denotes a column in data which includes the sum of a metric."""

SUM_SQ_COL = "sum_sq"
"""Denotes a column in data which includes the sum of a metric squared."""

CI_COL = "ci"
"""Denotes a column in data which includes the confidence interval of a metric."""

CI_PERCENT_COL = "ci_percent"
"""Denotes a column in data which includes the confidence interval of a metric as a percentage."""

DELTA_SUM_CI_COL = "delta_sum_ci"
"""Denotes a column in data which includes the confidence interval for difference of sum of metric between treatment and control."""

CATEG_NAN_VALUE = "nan"
"""Denotes a value in a categorical column which represents a missing value.
This is used to overwrite ad-hoc missing values in categorical columns,
which are usually a result of joins.
"""

CONTROL_LABEL = "control"
"""The typical label for the control group in an experiment."""

TREATMENT_LABEL = "treatment"
"""The typical label for the treatment group in an experiment."""

UNIT_COL = "memberid"
"""This is to specify the experiment unit column name for experiment assignments."""


EXPT_ID_COL = "exptid"
"""This is to specify the experiment ID column."""


METRIC_NAME_COL = "metric"
"""
Denotes a column in data which includes the metric name.
"""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LAW, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from dataclasses import dataclass

# These are the deduping methods
UNDEDUPED = "undeduped"
CONSECUTIVE_DEDUPED = "consecutive_deduped"
FULLY_DEDUPED = "fully_deduped"
MAP = "map"

DEDUPING_MTHODS = [UNDEDUPED, CONSECUTIVE_DEDUPED, FULLY_DEDUPED, MAP]


@dataclass
class SeqInfo:
    deduping_method: str = ""
    output_table_name: str = ""
    max_seq_index: int = 5
    """The max seq element number used in sunburst plots"""


# Now we define a useful constant to cover all sequence types we may need.
SEQ_INFO_LIST = []

for method in DEDUPING_MTHODS:
    SEQ_INFO_LIST.append(SeqInfo(deduping_method=method))



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""The goal is to simulate experiment data."""

from abvelocity.sim.sim import Sim

ATTRIBUTE_WEIGHTS = {
    "Level": {"Senior": 0.5, "Junior": 0.5},
    "Country": {"Iran": 0.1, "Canada": 0.3, "UK": 0.2, "US": 0.4},
    "Device": {"Smartphone": 0.6, "Tablet": 0.3, "Laptop": 0.1},
}
"""Example Define weights for attributes of the population."""


METRIC_ATTRIBUT_VALUES = {
    "metric1": {"Level": {"Junior": 5, "Senior": 2}, "Country": {"Iran": +5}},
    "metric2": {"Country": {"USA": 5, "Canada": 2}},
}
"""Example baseline metrics which depends on the attributes (but not experiment assignments)."""


POPULATION_SIZE = 1000
"""Default population size."""


POPULATION_SEED = 42
"""Default population seed."""


def simulate_data_multi1(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17]),
    population_pcnt_multi=tuple([100, 100]),
    non_trigger_pct_multi=tuple([5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    This has expt effects for two experiments on two metrics.
    But the interaction effect is only on metric1.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
    ]

    # Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: v2 is good. control is good.
    expt_metric_impacts = [
        {
            "control": {"metric1": 0, "metric2": 0},
            "v1": {"metric1": -2, "metric2": 0},
            "v2": {"metric1": +5, "metric2": +5},
        },  # Experiment 1 impact
        {
            "control": {"metric1": 0, "metric2": 0},
            "enabled": {"metric1": -2, "metric2": +1},
        },  # Experiment 2 impact
    ]

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled"): {"metric1": 15},
        ("v2", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 7
    assert sim.expt_metric_df.shape[1] == 9

    return sim


def simulate_data_multi2(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17]),
    population_pcnt_multi=tuple([90, 90]),
    non_trigger_pct_multi=tuple([5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    This has only experiment and interaction effects for metric1.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
    ]

    # Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: no impact
    expt_metric_impacts = [
        {
            "control": {"metric1": 1.5},
            "v1": {"metric1": -2, "metric2": 0},
            "v2": {"metric1": +5},
        },  # Experiment 1 impact
        {
            "control": {"metric2": 0},
            "enabled": {"metric1": -2, "metric2": 0},
        },  # Experiment 2 impact
    ]

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled"): {"metric1": 15},
        ("v2", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 7
    assert sim.expt_metric_df.shape[1] == 9

    return sim


def simulate_data_multi3(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17]),
    population_pcnt_multi=tuple([90, 90]),
    non_trigger_pct_multi=tuple([5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    In this case the effects come only through interactions and on metric1.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
    ]

    # No Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: no impact
    expt_metric_impacts = None

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled"): {"metric1": 15},
        ("v2", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 7
    assert sim.expt_metric_df.shape[1] == 9
    return sim


def simulate_data_multi4(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17]),
    population_pcnt_multi=tuple([90, 90]),
    non_trigger_pct_multi=tuple([5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    In this case the experiment assignments in each experiment are even.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 1.0 / 3.0, "v1": 1.0 / 3.0, "v2": 1.0 / 3.0},  # Experiment 1
        {"control": 0.5, "enabled": 0.5},  # Experiment 2
    ]

    # No Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: no impact
    expt_metric_impacts = None

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled"): {"metric1": 15},
        ("v2", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 7
    assert sim.expt_metric_df.shape[1] == 9
    return sim


def simulate_data_multi5(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17]),
    population_pcnt_multi=tuple([90, 90]),
    non_trigger_pct_multi=tuple([5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    In this case the experiment assignments in each experiment are even and each
    only have two arms 50/50 split.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.5, "v1": 0.5},  # Experiment 1
        {"control": 0.5, "enabled": 0.5},  # Experiment 2
    ]

    # No Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: no impact
    expt_metric_impacts = None

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {("v1", "enabled"): {"metric1": 15}}

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 7
    assert sim.expt_metric_df.shape[1] == 9
    return sim


def simulate_data_uni1(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13]),
    population_pcnt_multi=tuple([90]),
    non_trigger_pct_multi=tuple([5]),
) -> type(Sim):
    """
    An example with one experiment only.
    Only metric1 is impacted.
    """

    # Example parameters for univar experiment assignment weights
    # The assignment weights differ across variants.
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1 (the only experiment)
    ]

    # Define metrics impacts
    # Only metric 1 is impacted
    expt_metric_impacts = [
        {"control": {"metric1": 1.5}, "v1": {"metric1": -2, "metric2": 0}, "v2": {"metric1": +5}},
    ]

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=None,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 6
    assert sim.expt_metric_df.shape[1] == 8

    return sim


def simulate_data_uni2(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13]),
    population_pcnt_multi=tuple([90]),
    non_trigger_pct_multi=tuple([5]),
) -> type(Sim):
    """
    An example with one experiment only.
    Only metric1 is impacted.
    The population is split equally for three variants.
    """

    # Example parameters for univar experiment assignment weights
    # This time assignment weights are equal among the three variants.
    expt_variant_weights_multi = [
        {
            "control": 1.0 / 3.0,
            "v1": 1.0 / 3.0,
            "v2": 1.0 / 3.0,
        },  # Experiment 1 (the only experiment)
    ]

    # Define metrics impacts
    # Only metric 1 is impacted
    expt_metric_impacts = [
        {"control": {"metric1": 1.5}, "v1": {"metric1": -2}, "v2": {"metric1": +5}},
    ]

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=None,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 6
    assert sim.expt_metric_df.shape[1] == 8

    return sim


def simulate_data_uni3(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13]),
    population_pcnt_multi=tuple([90]),
    non_trigger_pct_multi=tuple([5]),
) -> type(Sim):
    """
    An example with one experiment only.
    Only metric1 is impacted.
    The population is split equally for only two variants (control and enabled).
    """

    # Example parameters for univar experiment assignment weights
    # This time assignment weights are equal among the three variants.
    expt_variant_weights_multi = [
        {"control": 1.0 / 2.0, "enabled": 1.0 / 2.0},  # Experiment 1 (the only experiment)
    ]

    # Define metrics impacts
    # Only metric 1 is impacted
    expt_metric_impacts = [
        {"control": {"metric1": 1.5}, "enabled": {"metric1": -2}},
    ]

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=None,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 6
    assert sim.expt_metric_df.shape[1] == 8

    return sim


def simulate_data_three1(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17, 19]),
    population_pcnt_multi=tuple([90, 90, 90]),
    non_trigger_pct_multi=tuple([5, 5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    This has expt effects for three experiments on two metrics.
    But the interaction effect is only on metric1.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
        {"control": 0.3, "enabled": 0.7},  # Experiment 3
    ]

    # Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: v2 is good. control is good.
    expt_metric_impacts = [
        {
            "control": {"metric1": 1.5},
            "v1": {"metric1": -2, "metric2": +5},
            "v2": {"metric1": +5},
        },  # Experiment 1 impact
        {
            "control": {"metric2": 1},
            "enabled": {"metric1": -2, "metric2": -5},
        },  # Experiment 2 impact
        {},  # Experiment 3, no impact
    ]

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled", "enabled"): {"metric1": 25},
        ("v2", "control", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 8
    assert sim.expt_metric_df.shape[1] == 10

    return sim


def simulate_data_three2(
    population_size=POPULATION_SIZE,
    population_seed=POPULATION_SEED,
    expt_assignment_seed_multi=tuple([13, 17, 19]),
    population_pcnt_multi=tuple([90, 90, 90]),
    non_trigger_pct_multi=tuple([5, 5, 5]),
) -> type(Sim):
    """
    This function creates simulated data to be used in tests.
    This has expt effects for three experiments on two metrics.
    But the interaction effect is only on metric1.
    """
    # Example parameters for multiple experiment assignment weights
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
        {"control": 0.3, "enabled": 0.7},  # Experiment 3
    ]

    # Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: v2 is good. control is good.
    expt_metric_impacts = [
        {
            "control": {"metric1": 1.5},
            "v1": {"metric1": -2, "metric2": 0},
            "v2": {"metric1": +5},
        },  # Experiment 1 impact
        {
            "control": {"metric2": 0},
            "enabled": {"metric1": -2, "metric2": 0},
        },  # Experiment 2 impact
        {},  # Experiment 3, no impact
    ]

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1
    interaction_metric_impacts = {
        ("v1", "enabled", "enabled"): {"metric1": 25},
        ("v2", "control", "control"): {"metric1": -2},
    }

    sim = Sim(
        population_size=population_size,
        attribute_weights=ATTRIBUTE_WEIGHTS,
        metric_attribute_values=METRIC_ATTRIBUT_VALUES,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=population_pcnt_multi,
        non_trigger_pct_multi=non_trigger_pct_multi,
        population_seed=population_seed,
        expt_assignment_seed_multi=expt_assignment_seed_multi,
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == population_size
    assert sim.expt_df.shape[1] == 8
    assert sim.expt_metric_df.shape[1] == 10

    return sim



# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# author: Reza Hosseini

from typing import Dict, Optional

import plotly.graph_objects as go
import plotly.io as pio


def gen_combined_figs_html(
    fig_dict: Dict[str, go.Figure], html_file_name: Optional[str] = None
) -> str:
    """
    Combines multiple Plotly figures into a single HTML string with captions and spacing.

    Args:
        fig_dict (Dict[str, go.Figure]): Dictionary where keys are figure names and values are Plotly figures.
        html_file_name (Optional[str]): Path to save the combined HTML file. If None, the result is not saved.

    Returns:
        str: The combined HTML string containing all figures with captions and spacing.
    """
    html_str = ""

    # Loop through the dictionary and create HTML for each figure with a caption and spacing
    for fig_name, fig in fig_dict.items():
        fig_html = pio.to_html(fig, full_html=False)
        # Adding a caption with the figure name and some spacing
        html_str += f"<h3>{fig_name}</h3><div style='margin-bottom: 30px;'>{fig_html}</div>"

    # Wrap the combined HTML in a basic HTML structure
    html_str = f"<html><body>{html_str}</body></html>"

    # Optionally save to file
    if html_file_name:
        with open(html_file_name, "w") as f:
            f.write(html_str)

    return html_str



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""The goal is to merge various experiment assignment DataContainers."""

from abvelocity.get_data.data_container import DataContainer
from abvelocity.param.constants import (
    CATEG_NAN_VALUE,
    TRIGGER_TIME_COL,
    UNIT_COL,
    VARIANT_COL,
)
from abvelocity.utils.merge_dfs_number_cols import merge_dfs_number_cols
from abvelocity.utils.merge_queries_number_cols import merge_queries_number_cols

ON_COLS = [UNIT_COL]
"""This is the expected list of columns which experiment data is to be joined on."""

COMMON_COLS = [VARIANT_COL, TRIGGER_TIME_COL]
"""These are the typical common columns across experiment data."""


def join_expt_dfs(
    dc_list: list[DataContainer],
    on_cols: list[str] = ON_COLS,
    common_cols: list[str] = COMMON_COLS,
    how: str = "outer",
    add_tuple: bool = True,
    drop_numbered_cols: bool = False,
) -> DataContainer:
    """Joins all experiment DataContainers by `UNIT_COL`.
    Typically we expect the common columns to be

        - VARIANT_COL
        - TRIGGER_TIME_COL

    Args:
        dc_list: A list of DataContainer objects containing experiment data.
        on_cols: Columns to join on.
        common_cols: Common columns across DataFrames.
        how: Type of merge to be performed.
        add_tuple: If True, add a tuple column.
        drop_numbered_cols: If True, drop numbered columns.

    Returns:
        A DataContainer containing the joined DataFrame.
    """

    # There will be three cases
    # Case 1: There are dfs materialized in all of the dcs
    # Case 2: dc_list contains SQL tables names
    # Case 3: dc_list contains SQL tables
    # To determine these we only assess the first element in dc_list
    # If it matches the case, we assume other elements do match as well

    if dc_list[0].df is not None:  # Case 1
        df_list = [dc.df for dc in dc_list if dc.df is not None]

        joined_df = merge_dfs_number_cols(
            df_list=df_list,
            on_cols=on_cols,
            common_cols=common_cols,
            how=how,
            add_tuple=add_tuple,
            nan_replacement=CATEG_NAN_VALUE,
            drop_numbered_cols=drop_numbered_cols,
        )

        return DataContainer(df=joined_df, is_df=True)
    elif dc_list[0].table_name is not None:  # Case 2
        query_list = [dc.table_name for dc in dc_list if dc.table_name is not None]
    else:  # Case 3
        query_list = [dc.query for dc in dc_list if dc.query is not None]

    # We have not returned for Case 2 and Case 3 and require more steps
    join_query = merge_queries_number_cols(
        queries=query_list,
        on_cols=on_cols,
        common_cols=common_cols,
        how=how,
        add_tuple=add_tuple,
        nan_replacements=[CATEG_NAN_VALUE, CATEG_NAN_VALUE],
        drop_numbered_cols=drop_numbered_cols,
    )

    return DataContainer(query=join_query, is_df=False)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""Module introduces a function to create a where clause for subsampling data."""

from typing import Optional


def subsample_condition(
    unit_col: str, mods: Optional[list[int]] = None, percent: Optional[int] = None
) -> Optional[str]:
    """Generates a SQL WHERE Clause based on `unit_col` which is assumed to
    be an int / bigint column.
    This is helpful to be able to subsample predictably across datasets.

    Args:
        unit_col: The column of data we are going to subsample based off.
        mods: A list of integers between 0 to 99 which to subsample based off.
            For example if

                - `mods = [1, 3]`
                - then only units in `unit_col` with remainders of 1 and 3
                when dividing by 100 will be accepted.

            If None, then `percent` will be used to construct mods.
        percent: An integer from 0 to 100 and determines the number of mods.
            For example if `percent = 3` then mods will be `[0, 1, 2]`.

    Returns:
        where_clause: The WHERE Clause for SQL.
            Assuming `unit_col = "id"`, a string of the form:

                - MOD(id, 100) IN (1, 2, 13)

            If inputs are None, then it returns None.

    Raises:
        None.
    """
    # If both mods and percent are None,
    # then None is returned.
    if mods is None and percent is None:
        return None

    if percent is not None:
        # Since mods start from zero, `range` works (as it also starts form zero).
        mods = range(int(percent))

    mods_string = """(""" + """,""".join([str(mod) for mod in mods]) + """)"""
    return f"""MOD({unit_col}, 100) IN {mods_string}"""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from dataclasses import dataclass
from typing import Optional


@dataclass
class UnivarStats:
    name: Optional[str] = None
    """The variable name."""
    mean: Optional[float] = None
    """The mean of the variable."""
    sd: Optional[float] = None
    """The standard deviation of the variable."""
    var: Optional[float] = None
    """The variance of the variable."""
    sample_count: Optional[float] = None
    """count of the variable in the sample."""
    sum: Optional[float] = None
    """sum of the variable."""
    sum_sq: Optional[float] = None
    """sum of the variable squared."""
    sample_mean_var: Optional[float] = None
    """This is an estimate of the variance of the sample mean.
    This field is useful especially in cases where the sample mean's variance is needed to
    be calculacuted in non-trivial ways, e.g., in the context of stratified sampling.
    In the simple case, the sample mean variance would be easily calculated as `var / count` and
    this field would be somewhat redundant.
    """
    triggered_count: Optional[float] = None
    """
    The triggered count refers to the count of all units,
    when a triggering mechanism is there and after triggering it assigns units to various
    alternatives (variants).
    In order to estimate the impact on the entire population, this value is utilized.
    For example it is used to calculate `.delta_sum` and a confidence interval  for it:
    `.delta_sum_ci` in `DeltaStats`.
    """


@dataclass
class TwoSampleTest:
    treatment_stats: UnivarStats
    """The test arm stats."""
    control_stats: UnivarStats
    """The control arm stats."""
    ci_coverage: float = 0.95
    """This denotes the coverage of the confidence interval."""
    same_impacted_population: bool = True
    """
    This bool determines if the impacted population by the two arms is the same.
    Note that in the experiment, there will be different number of triggers for
    each arm potentially in the sample (`sample_count`).
    But when this is True, we expect `.triggered_count` to be very close if not
    identical when this bool is True.
    """
    triggered_population_diff_thresh: float = 0.1
    """
    This is used to throw ValueError if `same_impacted_population` is True
    but the two arms contain different `.triggered_count` in them which are
    off by more than this threshold in terms of this metric: `abs(x-y) / (x+y)`,
    where x and y are the two different population.
    """


@dataclass
class DeltaStats:
    delta: Optional[float] = None
    """
    The raw difference between the two means of the populations.
    """
    delta_percent: Optional[float] = None
    """
    The percentage difference between the two means of the populations.
    """
    ci: Optional[float] = None
    """
    The confidence interval for the difference between the two means of the populations.
    """
    ci_percent: Optional[float] = None
    """
    The confidence interval for the percentage difference between the two means of the populations.
    """
    delta_std: Optional[float] = None
    """
    The standard deviation of the difference between the two means of the populations.
    """
    delta_sum: Optional[float] = None
    """
    The delta sum over the entire population (if population is finite).
    """
    delta_sum_ci: Optional[float] = None
    """
    The confidence interval for the difference in population sums.
    """
    z_value: Optional[float] = None
    """
    The Z-value for the difference between the two means of the populations.
    """
    p_value: Optional[float] = None
    """
    The P-value for the difference between the two means of the populations.
    """
    sample_counts: Optional[tuple[int]] = None
    """
    A tuple with:

        - The number of the units for the control arm.
        - The number of the units for the treatment arm.

    """
    impacted_counts: Optional[tuple[int]] = None
    """
    A tuple with:

    - The number of triggered (impacted) units across all variants
        with the same trigger state as control.
    - The number of triggered (impacted) units on all variants with the same
        trigger state as treatment.

    Note that these two values should be almost always either equal or very close.
    """


DELTA_STATS_OUTPUT_NAMES = ["delta", "delta(%)", "CI", "CI(%)", "delta_std", "z_value", "p_value"]
"""User-friendly output names for the delta stats.
These names are used in `two_sample_z_test` to generate the output dictionary.
TODO: Reza: Use these for creating user-friendly reports.
"""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import sqlite3
from typing import Any, List, Optional, Tuple

from abvelocity.get_data.cursor import Cursor


class SqliteCursor(Cursor):
    """
    A SQLite-specific cursor implementation for testing and compatibility with DataContainer.
    """

    def __init__(self, data_source: sqlite3.Connection):
        """
        Initializes the SqliteCursor with a SQLite connection.

        Args:
            data_source: A sqlite3.Connection object.
        """
        super().__init__(data_source)

    def execute(self, query: str) -> None:
        """
        Executes a SQLite query.

        Args:
            query: The SQL query string.
        """
        conn = self.data_source
        cursor = conn.cursor()
        try:
            cursor.execute(query)
            if query.strip().upper().startswith("SELECT"):
                self.description = cursor.description
                self.results = cursor.fetchall()
            else:
                conn.commit()  # Persist DDL/DML changes
                self.description = None
                self.results = None
        except sqlite3.Error as e:
            print(f"SQLite error: {e}")
            raise
        finally:
            cursor.close()

    def fetchall(self) -> Optional[List[Tuple[Any, ...]]]:
        """
        Fetches all rows of a query result as a list of tuples.

        Returns:
            A list of tuples representing the query results, or None if no results.
        """
        return self.results

    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        """
        Fetches the next row of a query result set.

        Returns:
            A tuple representing the next row, or None if no more rows.
        """
        if self.results:
            return self.results.pop(0) if len(self.results) > 0 else None
        return None

    def close(self) -> None:
        """Closes the SQLite connection."""
        if self.data_source:
            self.data_source.close()



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import re
from dataclasses import dataclass
from typing import Optional

DATE_COL_DEFAULT = "datepartition"


@dataclass
class EventTable:
    table_name: Optional[str] = None
    """SQL table name or a name to describe it in case `table_query` is passed below.
    This name is used to also construct the name of the output table: `output_table_name` as follows:
    `f"{create_table_prefix}_{convert_to_snake_case(event_table.table_name)}_{convert_to_snake_case(event_table.event_label)}"`
    """
    event_label: Optional[str] = None
    """Event label."""
    event_color: Optional[str] = None
    """Event color used in viz."""
    select_cols: Optional[list[str]] = None
    """List of column names to be selected in the query."""
    date_col: Optional[str] = None
    """Date column in the table.
    If not specified, it can be propagated from `common_info`.
    If still not set after propagation,
    it defaults to `DATE_COL_DEFAULT` in the query generation."""
    table_query: Optional[str] = None
    """This is the table query. Note that we allow queries
    rather than only relying on table_name in case some transormations are needed.
    If this is not passed we assume no transformation is needed and use `table_name`.
    """
    conditions: Optional[list[str]] = None
    """SQL conditions given as a list of strings."""
    output_table_name: Optional[str] = None
    """The name of output table created based on these specs (if / when created)"""
    start_date: Optional[str] = None
    """Start date for filtering records. If provided, it overrides the sequence start_date."""
    end_date: Optional[str] = None
    """End date for filtering records. If provided, it overrides the sequence end_date."""


@dataclass
class MultiEventTable:
    event_tables: list[EventTable]
    common_info: Optional[EventTable] = None

    def __post_init__(self):
        """
        Ensures that all event_labels in the event_tables list are unique.
        Raises a ValueError if duplicate event_labels are found.
        Propagation of common info is handled by the separate `propogate_common_info` method.
        """
        event_labels = []
        for event_table in self.event_tables:
            if event_table.event_label is not None:
                event_labels.append(event_table.event_label)

        # Check for uniqueness
        if len(event_labels) != len(set(event_labels)):
            duplicate_labels = {label for label in event_labels if event_labels.count(label) > 1}
            raise ValueError(
                f"All event_labels within 'event_tables' must be unique. Duplicates found: {duplicate_labels}"
            )

    def propogate_common_info(self):
        """propogates the info from `common_info` for missing fields of each event_table"""
        # We've removed the early `return` here so that defaults can be applied even if common_info is None.
        for event_table in self.event_tables:
            # Propagate table_name
            if (
                event_table.table_name is None
                and self.common_info is not None
                and self.common_info.table_name is not None
            ):
                event_table.table_name = self.common_info.table_name

            # Propagate select_cols
            if (
                event_table.select_cols is None
                and self.common_info is not None
                and self.common_info.select_cols is not None
            ):
                event_table.select_cols = self.common_info.select_cols

            # Propagate date_col
            if event_table.date_col is None:
                if self.common_info is not None and self.common_info.date_col is not None:
                    event_table.date_col = self.common_info.date_col
                else:
                    # If both the event table and common info have no date_col, use the default.
                    event_table.date_col = DATE_COL_DEFAULT

            # Propagate table_query
            if (
                event_table.table_query is None
                and self.common_info is not None
                and self.common_info.table_query is not None
            ):
                event_table.table_query = self.common_info.table_query

            # Propagate conditions: Add common conditions to existing ones if they exist
            if self.common_info is not None and self.common_info.conditions is not None:
                if event_table.conditions is None:
                    event_table.conditions = self.common_info.conditions
                else:
                    event_table.conditions = self.common_info.conditions + event_table.conditions

            # Propagate start_date
            if (
                event_table.start_date is None
                and self.common_info is not None
                and self.common_info.start_date is not None
            ):
                event_table.start_date = self.common_info.start_date

            # Propagate end_date
            if (
                event_table.end_date is None
                and self.common_info is not None
                and self.common_info.end_date is not None
            ):
                event_table.end_date = self.common_info.end_date

            # event_label, event_color, and output_table_name are typically specific
            # to each EventTable and thus not propagated from common_info.


def convert_to_snake_case(input_string: str) -> str:
    """
    Converts a given string to snake_case.
    Replaces periods with underscores and converts camel case to snake case.
    """
    # Helper function to replace all periods with underscores
    input_string = input_string.replace(".", "_")

    # Convert camel case to snake case
    # Insert an underscore before each uppercase letter that follows a lowercase letter
    input_string = re.sub(r"([a-z])([A-Z])", r"\1_\2", input_string)

    # Convert the string to lowercase and return
    return input_string.lower()


def gen_event_query(
    event_table: EventTable,
    start_date: str,
    end_date: str,
    create_table_prefix: Optional[str] = None,
) -> str:
    """Generates a SQL query to extract tracking event data from a specified table.
    This assumes a row in the target table is logging an occurrence of an event: `event_label`.
    The resulting created table name (if created) will be
    `f"{create_table_prefix}_{convert_to_snake_case(event_table.table_name)}_{convert_to_snake_case(event_table.event_label)}"`.

    Args:
        event_table (EventTable): Dataclass containing table details, event label, columns, and conditions.
        start_date (str): The start date for filtering records (format: 'YYYY-MM-DD-00'). This will be used if `event_table.start_date` is not provided.
        end_date (str): The end date for filtering records (format: 'YYYY-MM-DD-00'). This will be used if `event_table.end_date` is not provided.
        create_table_prefix (Optional[str], optional): If provided, a new table will be created with this prefix.

    Returns:
        str: The generated SQL query as a string.

    Raises:
        ValueError: If any required parameters (table_name or table_query, event_label, start_date, end_date) are empty or None.
    """
    table_start_date = event_table.start_date or start_date
    table_end_date = event_table.end_date or end_date

    # Use the date_col from the event table. At this point, it should be set
    # by user input, propagation from common_info, or the default.
    date_col_to_use = event_table.date_col

    # Validate that either table_name or table_query is provided
    if not (event_table.table_name or event_table.table_query):
        raise ValueError("Either 'table_name' or 'table_query' must be provided in EventTable.")

    if not all([event_table.event_label, table_start_date, table_end_date]):
        raise ValueError(
            "All required parameters (event_label, start_date, end_date) must be provided."
        )

    # Determine the source for the query (table_query takes precedence if both are present)
    query_table = (
        f"({event_table.table_query})" if event_table.table_query else event_table.table_name
    )

    # Select columns: if select_cols is None or empty, assume all columns (SELECT *)
    select_clause = ", ".join(event_table.select_cols) if event_table.select_cols else "*"

    query = f"""
        SELECT
            {select_clause},
            '{event_table.event_label}' AS event
        FROM {query_table}
        WHERE TRUE
            AND {date_col_to_use} BETWEEN '{table_start_date}' AND '{table_end_date}'"""

    if event_table.conditions:
        query += " AND " + " AND ".join(event_table.conditions)

    if create_table_prefix:
        # Ensure event_table.table_name is not None for output table naming if table_query is used
        # You might need a more robust naming convention if table_name is Optional and table_query is used.
        # A safer approach might involve using event_table.event_label for naming if table_name is absent.
        name_for_output = (
            event_table.table_name if event_table.table_name else "custom_query"
        )  # Placeholder if table_name is None but table_query is used

        output_table_name = f"{create_table_prefix}_{convert_to_snake_case(name_for_output)}_{convert_to_snake_case(event_table.event_label)}"
        event_table.output_table_name = output_table_name
        query_prefix = f"""
        DROP TABLE IF EXISTS {output_table_name};
        CREATE TABLE IF NOT EXISTS {output_table_name} AS"""
        query = query_prefix + query

    return query



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from typing import Any, Callable, Dict, List, Optional

import pandas as pd
import plotly.express as px
from plotly.graph_objects import Figure


def get_categ_column_distbn(
    table_name: str,
    col: str,
    count_distinct_col: Optional[str] = None,
    cursor: Optional[Any] = None,
    print_to_html: Optional[Callable] = None,
    max_num: int = 20,
    conditions: List[str] = ["TRUE = TRUE"],
    file_name: Optional[str] = None,
    title: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Retrieves and visualizes the distribution of a categorical column in a database table.

    This function executes a SQL query to count the occurrences of each unique value in a specified column,
    optionally filtering the results based on a given condition. It then generates a bar chart showing the
    distribution of the top 'max_num' values, along with an 'Other' category for the remaining values.

    Args:
        table_name (str): The name of the database table.
        col (str): The name of the categorical column to analyze.
        curosr (Any): SQL database cursor.
        count_distinct_col (str): The name of the column containing the distinct values to be counted.
        print_to_html (Optional[Callable], optional):
            A function to print output (text or figures) to an HTML file. Defaults to None.
            It should accept a 'message' and optionally 'color', 'font_size', and 'file_name' arguments.
        max_num (int, optional): The number of top values to display in the chart. Defaults to 20.
        conditions (List[str], optional): A list of SQL conditions to filter the data. Defaults to "TRUE = TRUE" (no filtering).
        file_name (str, optional): The name of the HTML file to write output to. Defaults to None.
        title (str, optional): If not passed, one will be generated based on `table_name`, `col`, `conditions` etc.

    Returns:
        Dict[str, Any]: A dictionary containing the following keys:
            - 'fig' (plotly.graph_objects.Figure): The generated bar chart figure.
            - 'df' (pandas.DataFrame): The full DataFrame containing the counts of all unique values.
            - 'plot_df' (pandas.DataFrame): The DataFrame used for plotting, including the top 'max_num' values and 'Other'.
            - 'num_distict_labels' (int): the number of unique values in the column.
    """
    # If col has the form `string1.string2` SQL will return that col as `string2`.
    # Therefore we define `col0` and use it.
    col0 = col.rsplit(".", 1)[1] if "." in col else col

    condition = " AND ".join(conditions)

    count_str = "COUNT(*) AS count"
    if count_distinct_col:
        count_str = f"COUNT(DISTINCT {count_distinct_col}) AS count"

    query = f"""
    SELECT
        COALESCE({col}, 'NA') AS {col0},
        {count_str}
    FROM
        {table_name}
    WHERE {condition}
    GROUP BY
        {col}
    ORDER BY
        count DESC
    """

    header = (
        f"Getting the distribution of possible values for {col} in {table_name}."
        + f" Counting rows for each value of {col}"
        + f" Conditions: {conditions}"
    )

    if count_distinct_col:
        header = (
            f"Getting the distribution of possible values for {col} in {table_name}."
            + f" Counting distinct {count_distinct_col} for each value of {col}"
            f" Conditions: {conditions}"
        )

    if print_to_html:
        print_to_html(
            header,
            file_name=file_name,
        )
        print_to_html(query, color="grey", font_size=8, file_name=file_name)

    if not cursor:
        return

    df_wrapper = cursor.get_df(query=query)
    df = df_wrapper.df

    if not len(df):
        print("df is empty. df is returning.")
        return

    num_distinct_labels = df[col0].nunique()
    if print_to_html:
        print_to_html(
            f"Total number of distinct labels for {col} across {table_name} is: {num_distinct_labels}",
            file_name=file_name,
        )

    top_k = df.head(max_num).copy()
    other_count = df["count"][max_num:].sum()
    total_count = df["count"].sum()

    if other_count > 0:
        other_row = pd.DataFrame([{col0: "Other", "count": other_count}])
        plot_df = pd.concat([top_k, other_row], ignore_index=True)
    else:
        plot_df = top_k

    plot_df["percent"] = 100.0 * plot_df["count"].copy() / total_count

    fig = px.bar(
        plot_df,
        x=col0,
        y="percent",
        title=f"Distribution of {col} in {table_name}: count_distinct_col: {count_distinct_col}, conditions: {conditions}",
    )

    if print_to_html:
        print_to_html(f"Number of distinct {col0}: {num_distinct_labels}", file_name=file_name)

        print_to_html(fig, file_name=file_name)

    return {
        "fig": fig,
        "df": df,
        "plot_df": plot_df,
        "query": query,
        "num_distict_labels": num_distinct_labels,
    }


def num_distinct_values_per_entity(
    table_name: str,
    entity_cols: List[str],
    count_distinct_col: str,
    cursor: Optional[Any] = None,
    print_to_html: Optional[Callable] = None,
    conditions: List[str] = ["TRUE=TRUE"],
    x_range: Optional[List[float]] = None,
    file_name: Optional[str] = None,
) -> Figure:
    """
    Calculates and visualizes the distribution of the number of distinct values per entity.

    This function executes a SQL query to count the number of distinct values in a specified
    column for each unique combination of entity columns. It then generates a histogram
    showing the distribution of these counts.

    Args:
        table_name (str): The name of the database table.
        entity_cols (List[str]): A list of column names representing the entity.
        count_distinct_col (str): The name of the column containing the distinct values to be counted.
        cursor (Optional[Any]): The database cursor object.
        print_to_html (Optional[Callable], optional): A function to print output (text or figures) to an HTML file.
            It should accept a 'message' and optionally 'color', 'font_size', and 'file_name' arguments.
        conditions (List[str], optional): A list of SQL condition clauses to filter the data. Defaults to `["TRUE=TRUE"]` (no filtering).
        x_range (Optional[List[float]], optional): A list specifying the x-axis range for the histogram. Defaults to None.
        file_name (Optional[str], optional): The name of the HTML file to write output to. Defaults to None.

    Returns:
        Figure: The generated histogram figure.
    """
    condition = " AND ".join(conditions)
    if print_to_html:
        print_to_html(
            f"Counts number of distinct {count_distinct_col} per {entity_cols} and displays the distribution.",
            file_name=file_name,
        )

    entity_cols_str = ",".join(entity_cols)

    query = f"""
    SELECT
        {entity_cols_str},
        COUNT(DISTINCT {count_distinct_col}) AS distinct_count
    FROM
        {table_name}
    WHERE {condition}
    GROUP BY
        {entity_cols_str}
    """

    if print_to_html:
        print_to_html(query, color="grey", font_size=8, file_name=file_name)

    df_wrapper = cursor.get_df(query=query)
    df = df_wrapper.df

    fig = px.histogram(
        df,
        x="distinct_count",
        title=f"Distribution of {count_distinct_col} per {entity_cols} in {table_name}: conditions: {conditions}",
        histnorm="percent",
    )

    if x_range:
        fig.update_layout(xaxis=dict(range=x_range))

    if print_to_html:
        print_to_html(fig, file_name=file_name)

    return {"fig": fig, "df": df, "query": query}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini


import os
from pathlib import Path

from sqlglot import transpile


class SqlConversion:
    """
    A class to handle SQL dialect conversions using sqlglot.

    The class is initialized with a source and a target dialect.
    It can then convert a SQL string from the source to the target.

    Valid dialects can be found in the sqlglot documentation, for example:
    'spark', 'trino', 'hive', 'presto', 'mysql', 'postgres', etc.
    """

    def __init__(self, from_dialect: str, to_dialect: str):
        """
        Initializes the SqlConversion instance.

        Args:
            from_dialect (str): The name of the source SQL dialect.
            to_dialect (str): The name of the target SQL dialect.
        """
        # Store the source and target dialects
        self.from_dialect = from_dialect
        self.to_dialect = to_dialect

    def convert(self, sql_query: str) -> str:
        """
        Transpiles a given SQL query from the source to the target dialect.

        Args:
            sql_query (str): The SQL query string to be converted.

        Returns:
            str: The converted SQL query string.
        """
        try:
            # Use sqlglot's transpile function to perform the conversion
            converted_query = transpile(
                sql_query,
                read=self.from_dialect,
                write=self.to_dialect,
                pretty=True,  # Optional: formats the output nicely
            )
            # The transpile function returns a list, so we take the first element.
            return converted_query[0]
        except Exception as e:
            # Simple error handling for cases where the SQL can't be parsed
            return f"Error during conversion: {e}"

    def convert_directory(self, input_path: str, output_path: str = None):
        """
        Recursively converts all SQL files in a directory and its subdirectories.

        Args:
            input_path (str): The path to the directory containing SQL files.
            output_path (str, optional): The path to the output directory
                for converted files. If None,
                files are saved in the same directory
                with a new filename. Defaults to None.
        """
        # Use Path for easy path manipulation and creation
        input_dir = Path(input_path)

        if not input_dir.is_dir():
            print(f"Error: The provided input path '{input_path}' is not a directory.")
            return

        if output_path:
            output_dir = Path(output_path)
            output_dir.mkdir(parents=True, exist_ok=True)
        else:
            output_dir = input_dir

        print(f"Starting conversion from '{self.from_dialect}' to '{self.to_dialect}'...")
        print(f"Input directory: {input_dir}")
        print(f"Output directory: {output_dir}")

        for root, dirs, files in os.walk(input_dir):
            for file in files:
                if file.endswith(".sql"):
                    # Construct the full path to the input file
                    input_file_path = Path(root) / file
                    print(f"Converting file: {input_file_path}")

                    # Read the SQL content
                    with open(input_file_path, "r") as f:
                        sql_content = f.read()

                    # Convert the content
                    converted_sql = self.convert(sql_content)

                    # Determine the output file path
                    if output_path:
                        # Recreate the subdirectory structure in the output path
                        relative_path = input_file_path.relative_to(input_dir)
                        output_file_path = output_dir / relative_path
                        output_file_path.parent.mkdir(parents=True, exist_ok=True)
                    else:
                        # Append the target dialect to the filename
                        filename_stem = input_file_path.stem
                        output_file_path = (
                            input_file_path.parent / f"{filename_stem}_{self.to_dialect}.sql"
                        )

                    # Write the converted content to the new file
                    with open(output_file_path, "w") as f:
                        f.write(converted_sql)

                    print(f"Saved converted query to: {output_file_path}")
        print("Conversion of all files completed.")



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import copy
from typing import Callable, Optional

from abvelocity.get_data.cursor import Cursor
from abvelocity.mea.get_mea_data import get_mea_data
from abvelocity.mea.mea import MEA, MEAResult
from abvelocity.param.analysis_info import AnalysisInfo
from abvelocity.param.launch import Launch
from abvelocity.param.variant import ComparisonPair


def run_mea(
    cursor: Cursor,
    analysis_info: AnalysisInfo,
    expt_asgmnt_table: str,
    get_asgmnt_query: Callable,
    launches: Optional[list[Launch]] = None,
    control_launch: Optional[Launch] = None,
    comparison_pairs: list[ComparisonPair] = None,
    scale: bool = True,
    condition: Optional[str] = None,
) -> Optional[MEAResult]:
    """This is the flow to run multi-experiment analysis (MEA).
    This takes two steps:

        - Get the data for MEA.
        - Run MEA.

    Args:
        cursor: A database Cursor.
        analysis_info: AnalysisInfo which includes the experiments and metrics.
        expt_asgmnt_table: Table which includes expt data.
        get_asgmnt_query: Callable which creates expt query.
        launches: List of launches. Each launch is a combination of variants across experiments (or simply a string for single experiment case).
            See `~abvelocity.param.launch.Launch`.
            For each launch first,

                - we construct the counterpart multi-experiment control
                - for each arm we map them to the corresponding `VariantList`.
                - a `ComparisonPair` is created with these two objects then.

            All the above steps are done using `launch_to_comparison_pair` function.
        control_launch: The Launch which is used as the baseline for comparison.
            If not passed, we will assume all experiments are on the control arm (`CONTROL_LABEL`).
        comparison_pairs: List of comparison pairs.
            This is an advanced feature as `launches` will construct the needed comparison_pairs for typical launches.
            Each pair includes a treatment and control field.
            See `~abvelocity.param.variant.ComparisonPair`.
        scale: If true and metric_info_list has more than one element,

            - we get the assignment data `expt_df` and
            - for each group (can think of this as map phase):
                - join with expt_df
                - perform mea
                - store results
                - delete joined data
            - combine the results (can think of this as reduce phase)
        condition: Optional SQL query condition to be passed to `get_mea_data`

    Returns:
        An instance of `MEAResult`.
        In summary, the result will include the effect sizes and confidence intervals.
        See `~abvelocity.mea.mea.MEAResult` for more details.

    Alters:
        analysis_info:

            - This will be altered upon reading the data and some
                statistics will be added regarding experiments.
            - Also if no metric is passed via `analysis_info`, an attempt
                will be made to infer metrics from `metric_family_name`.


    """
    # Gets the data needed for MEA and updates `analysis_info` by attaching
    # derived expt statistics and metric, metrics are not passed directly
    # in metrics field.
    if scale and analysis_info.metric_info_list and len(analysis_info.metric_info_list) > 1:
        print(
            f"\n*** scale is true and there are more than 1 metric groups: {len(analysis_info.metric_info_list)}"
        )
        print("\n*** we will join metric groups in various stages to limit memory usage.")

        # We start from an empty mea result and keep augmenting it in the for loop.
        mea_result = MEAResult()

        analysis_info_copy = copy.deepcopy(analysis_info)
        analysis_info_copy.metric_info_list = None
        expt_dc = get_mea_data(
            cursor=cursor,
            analysis_info=analysis_info_copy,
            expt_asgmnt_table=expt_asgmnt_table,
            get_asgmnt_query=get_asgmnt_query,
            condition=condition,
        )
        print(f"*** expt_dc was obtained by passing an analysis_info w/o : {expt_dc}")
        # We want to update the `derived_stats` for `analysis_info.multi_expt_info`
        # This is despite passing a copy which is merely done to remove the metric info.
        analysis_info.multi_expt_info.derived_stats = (
            analysis_info_copy.multi_expt_info.derived_stats
        )

        # We also want to update the `start_date` `end_date` of `analysis_info`
        analysis_info.start_date = analysis_info_copy.start_date
        analysis_info.end_date = analysis_info_copy.end_date

        for metric_info in analysis_info.metric_info_list:
            # We reset the `metric_info_list` to only include one `metric_info`
            print(f"\n*** running MEA for metric_info:\n {metric_info}")
            analysis_info_copy.metric_info_list = [metric_info]
            df = get_mea_data(
                cursor=cursor,
                analysis_info=analysis_info_copy,
                expt_asgmnt_table=expt_asgmnt_table,
                get_asgmnt_query=get_asgmnt_query,
                existing_dc=expt_dc,
                condition=condition,
            ).df
            print(f"\n*** MEA data was obtained: df\n {df}")
            mea = MEA(
                df=df,
                analysis_info=analysis_info_copy,
                launches=launches,
                control_launch=control_launch,
                comparison_pairs=comparison_pairs,
            )

            mea.run()
            print("\n*** Delete df generated for this run of MEA.")
            del df

            mea_result_new = mea.result
            print("\n*** Deletes `MEA` object after extracting the result.")
            del mea
            mea_result.combine(mea_result_new)
    else:
        # This will also handle the case with no metrics
        # In such case mea_result will be None
        # However `analysis_info` could get updated with experiment stats
        df = get_mea_data(
            cursor=cursor,
            analysis_info=analysis_info,
            expt_asgmnt_table=expt_asgmnt_table,
            get_asgmnt_query=get_asgmnt_query,
            condition=condition,
        ).df

        mea = MEA(
            df=df,
            analysis_info=analysis_info,
            launches=launches,
            control_launch=control_launch,
            comparison_pairs=comparison_pairs,
        )

        mea.run()

        mea_result = mea.result
        del df
        del mea

    return mea_result



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import math

import numpy as np
from scipy import stats

from abvelocity.stats.stats import DeltaStats, TwoSampleTest

SMALL_SAMPLE_SIZE = 5
"""A constant to handle corner cases where the sample is too small."""


def calc_standard_normal_ci(mean: float, sd: float, ci_coverage: float) -> dict:
    """
    Calculates the z-value, p-value, and confidence interval (CI) for a standard normal distribution.
    Note that here `sd` is the standard deviation for the sample mean and that is why no sample size is given.
    This is useful when `sd` can be estimated for more complex situations e.g. when the sample mean is estimated
    for weighted samples.

    Args:
        mean: The mean of the distribution.
        sd: The standard deviation of the distribution. Must be non-negative.
        ci_coverage: The desired coverage of the confidence interval, a value between 0 and 1.

    Returns:
        dict: A dictionary containing the following keys:
            - "z_value" (float): The z-value, which is the number of standard deviations the mean is from zero.
            - "p_value" (float): The two-tailed p-value corresponding to the z-value.
            - "ci" (numpy.ndarray): A 1D array with two elements, representing the lower and upper bounds of the confidence interval.

    Raises:
        ValueError: If `sd` is negative.
        ValueError: If `ci_coverage` is not between 0 and 1.

    """
    if sd < 0:
        raise ValueError(f"sd has to be non-negative: {sd}")

    # If sd is None, we return
    if not sd or math.isnan(sd):
        return {"z_value": np.nan, "p_value": np.nan, "ci": np.array([np.nan, np.nan])}

    if not (0 < ci_coverage < 1):
        raise ValueError(f"ci_coverage has to be between 0 and 1: {ci_coverage}")

    # If both sd and mean are zero, we do not divide 0/0 and assign a signifcant z_value of magnitude 5
    # If only sd is zero z_value will be 5 times the sign of the mean
    if sd == 0 and mean == 0:
        z_value = 5.0
    elif sd == 0:
        z_value = np.sign(mean) * 5.0
    else:
        z_value = mean / sd

    p_value = 2 * (1 - stats.norm.cdf(abs(z_value), loc=0, scale=1))
    # Calculates Z (standard normal) quantile for a given confidence interval coverage level
    tail_mass = 1 - (1 - ci_coverage) / 2
    # Calculate standard normal quantile for that tail
    ci_radius_coef = stats.norm.ppf(tail_mass, loc=0, scale=1)
    ci_radius = ci_radius_coef * sd
    ci = np.array([mean - ci_radius, mean + ci_radius])

    return {"z_value": z_value, "p_value": p_value, "ci": ci}


def two_sample_z_test(two_sample_test: TwoSampleTest) -> DeltaStats:
    """Z test to compare two populations with their univariate stats given in:

        - two_sample_test["treatment_stats"] (treatment arm)
        - two_sample_test["control_stats"] (control arm)

    This assumes the sample mean is available for both arms in `.mean` attribute.

    Then there are two cases for sample mean variance for each arm:

        - (a) `.sample_mean_var` is already available.
        - (b) `.sample_mean_var` is None. But `.var` and `.sample_count` is available.

    For (a) the test will directly use the sample mean variance to calculate confidence intervals
    and p-values etc.
    For (b), sample mean var will be computed here by dividing `.var` by `.sample_count` attributes.

    It uses pre-calculated means, variances
    and sample sizes (sample_count) from the two samples.

    The assumptions are

        - the two samples are independent of each other
        - each sample consists of independent and identically distributed observations

    Args:
        two_sample_test: A dataclass which includes:

            - the univariate stats for the treatment and control arms
            - the confidence interval coverage.

    Returns:
        delta_stats: A dataclass which includes experiment results e.g. delta, CI, p-value etc.

    Raises:
        None.
    """
    # First for each arm, if `.sample_mean_var` is not available,
    # we calculate it by dividing `.var` by `.sample_count`
    for arm in [two_sample_test.treatment_stats, two_sample_test.control_stats]:
        if arm.sample_mean_var is None:
            if arm.var is None or arm.sample_count is None:
                raise ValueError(
                    f"if `.sample_mean_var` is missing `.var` and `.sample_count` need to be available in {arm}"
                )
            else:
                arm.sample_mean_var = arm.var / arm.sample_count

        # We also check for existence of .mean in both arms
        if arm.mean is None:
            raise ValueError(f"`.mean` cannot be None for an experiment arm: {arm}")

    treatment_stats = two_sample_test.treatment_stats
    control_stats = two_sample_test.control_stats
    ci_coverage = two_sample_test.ci_coverage

    delta = treatment_stats.mean - control_stats.mean

    t_var = treatment_stats.sample_mean_var
    c_var = control_stats.sample_mean_var

    delta_var = 0.0

    # These following operations are to handle rare corner cases
    # Check for NaN cases first, as NaN propagates and needs explicit handling
    if math.isnan(t_var) and math.isnan(c_var):
        delta_var = float("nan")
    # If one of them is Nan we use a conservative variance
    # by multiplying the existing one by 3
    # If both are not Nan, we add them (this is the default situation)
    elif math.isnan(t_var):
        delta_var = c_var * 3
    elif math.isnan(c_var):
        delta_var = t_var * 3
    else:
        delta_var = t_var + c_var

    delta_std = np.sqrt(delta_var)

    z_value = np.nan
    p_value = np.nan
    ci = np.array([np.nan, np.nan])  # Initialize CI to NaN as well for consistency

    if (
        control_stats.sample_count is not None and control_stats.sample_count <= SMALL_SAMPLE_SIZE
    ) or (
        treatment_stats.sample_count is not None
        and treatment_stats.sample_count <= SMALL_SAMPLE_SIZE
    ):
        # If either sample count is less than 4, return NaNs for z_value and p_value.
        # ci will remain NaN from initialization.
        pass  # z_value, p_value, and ci are already initialized to NaN
    else:
        # Only calculate if sample counts are sufficient
        sig_res = calc_standard_normal_ci(mean=delta, sd=delta_std, ci_coverage=ci_coverage)
        z_value = sig_res["z_value"]
        p_value = sig_res["p_value"]
        ci = sig_res["ci"]

    # TODO: Reza, the CI below can be improved by accounting for the randomness in the mean in denominator.
    # This is not a major, issue though as sample sizes are typically large for the control arm.
    # We need to handle potential division by zero if control_stats.mean is 0
    if control_stats.mean != 0:
        delta_percent = round(100 * delta / control_stats.mean, 3)
        ci_percent = (100 * ci / control_stats.mean).round(3)
    else:
        delta_percent = np.nan
        ci_percent = np.array([np.nan, np.nan])

    # Calculate the difference between sum of metric in treatment and control arms.
    # This includes calculating conf. interval for the difference in the sums.
    delta_sum = None
    delta_sum_ci = None
    t_triggered_count = treatment_stats.triggered_count
    c_triggered_count = control_stats.triggered_count
    if t_triggered_count is not None and c_triggered_count is not None:
        if two_sample_test.same_impacted_population:
            # In this case we assume the impacted population is the same.
            # This implies `.triggered_count` to be very close if not identical.
            # Therefore we take their average and will issue a warning if they
            # are not close enough.
            triggered_count = 1 / 2 * (t_triggered_count + c_triggered_count)
            diff = abs(t_triggered_count - c_triggered_count) / (
                t_triggered_count + c_triggered_count
            )

            if diff > two_sample_test.triggered_population_diff_thresh:
                raise ValueError(
                    "`same_impacted_population` is True, yet the `.triggered_count` are very different on the two arms: "
                    f"treatment: {t_triggered_count}, control: {c_triggered_count}"
                )

            delta_sum = (
                triggered_count * treatment_stats.mean - triggered_count * control_stats.mean
            )
            # To calculate variance, note that we need a power of two for constant multipliers
            # Also note that it is summable since the control and treament are assumed to be independent
            delta_sum_var = (triggered_count**2) * treatment_stats.sample_mean_var + (
                triggered_count**2
            ) * control_stats.sample_mean_var

        else:
            delta_sum = (
                t_triggered_count * treatment_stats.mean - c_triggered_count * control_stats.mean
            )
            # To calculate variance, note that we need a power of two for constant multipliers
            # Also note that it is summable since the control and treament are assumed to be independent
            delta_sum_var = (t_triggered_count**2) * treatment_stats.sample_mean_var + (
                c_triggered_count**2
            ) * control_stats.sample_mean_var

        delta_sum_std = np.sqrt(delta_sum_var)
        # Apply the same sample count check for delta_sum_ci as well
        if (
            control_stats.sample_count is not None
            and control_stats.sample_count <= SMALL_SAMPLE_SIZE
        ) or (
            treatment_stats.sample_count is not None
            and treatment_stats.sample_count <= SMALL_SAMPLE_SIZE
        ):
            delta_sum_ci = np.array([np.nan, np.nan])
        else:
            sig_res_sum = calc_standard_normal_ci(
                mean=delta_sum, sd=delta_sum_std, ci_coverage=ci_coverage
            )
            delta_sum_ci = sig_res_sum["ci"]

    return DeltaStats(
        delta=delta,
        delta_percent=delta_percent,
        ci=ci,
        ci_percent=ci_percent,
        delta_std=delta_std,
        delta_sum=delta_sum,
        delta_sum_ci=delta_sum_ci,
        z_value=z_value,
        p_value=p_value,
        sample_counts=(control_stats.sample_count, treatment_stats.sample_count),
        impacted_counts=(control_stats.triggered_count, treatment_stats.triggered_count),
    )



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

"""The goal is to get "Expt" data which is experiment assignment data.
It also joins several expts if needed."""

import sys
from typing import Callable, List, Optional

from abvelocity.get_data.cursor import Cursor
from abvelocity.get_data.data_container import DataContainer
from abvelocity.get_data.get_expt_stats import get_expt_stats
from abvelocity.param.constants import UNIT_COL
from abvelocity.param.expt_info import ExptInfo, MultiExptInfo


def get_asgmnt_data(
    expt_asgmnt_table: str,
    get_asgmnt_query: Callable,
    expt_info: ExptInfo,
    cursor: Optional[Cursor] = None,
    condition: Optional[str] = None,
    materialize_to_df: bool = True,
) -> DataContainer:
    """Returns experiment assignment data as a DataContainer.

    Args:
        expt_asgmnt_table: The experiment assignment table.
        get_asgmnt_query: Callable
        expt_info: Experiment info.
            See `~abvelocity.expt_info.ExptInfo` for more details.
        cursor: A Presto Cursor.
        condition: A SQL condition.
        materialize_to_df: A bool to determine if we want to materialize to pandas df.

    Returns:
        result: DataContainer with experiment assignment data.

    Raises:
        None.
    """
    query = expt_info.query
    if query is None:
        query = get_asgmnt_query(
            expt_asgmnt_table=expt_asgmnt_table, expt_info=expt_info, condition=condition
        )

    if not materialize_to_df:
        return DataContainer(query=query, is_df=False)

    result = cursor.get_df(query=query)

    return DataContainer(df=result.df, is_df=True)


def get_asgmnt_data_multi(
    expt_asgmnt_table: str,
    get_asgmnt_query: Callable,
    multi_expt_info: MultiExptInfo,
    cursor: Optional[Cursor] = None,
    condition: Optional[str] = None,
    drop_unit_duplicates: bool = True,
    materialize_to_df: bool = True,
) -> List[DataContainer]:
    """This function queries expt data and returns a list of DataContainers.

    Args:
        expt_asgmnt_table : `str`
            The directory to query expt data.
        get_asgmnt_query: Callable
        multi_expt_info : `MultiExptInfo`
            A MultiExptInfo object with experiment information.
        cursor: A databse Cursor.
        condition: A SQL condition.
        drop_unit_duplicates : If True, this will only keep one row per unit (`UNIT_COL`) for each experiment.
            It keeps the last row, but note that this is a rather arbitrary choice.
            Best practice is to pass data without duplicates in `UNIT_COL` to this function.
            This argument is only supported when we do materialize to df.
        materialize_to_df: A bool to determine if we want to materialize to pandas df.

    Returns
        expt_dc_list : A list of DataContainers, one for each experiment in `multi_expt_info.expt_info_list`.

    Alters:
        expt_info_list : This function will amend each `expt_info` in the list with
        it's derived stats (`DerivedExptState`)
    """

    # Queries expt data for each expt in `multi_expt_info.expt_info_list`.
    expt_dc_list = []
    for expt_info in multi_expt_info.expt_info_list:
        dc = get_asgmnt_data(
            cursor=cursor,
            expt_asgmnt_table=expt_asgmnt_table,
            get_asgmnt_query=get_asgmnt_query,
            expt_info=expt_info,
            condition=condition,
            materialize_to_df=materialize_to_df,
        )
        if materialize_to_df:
            expt_df = dc.df

            if drop_unit_duplicates:
                expt_df.drop_duplicates(subset=[UNIT_COL], keep="first", inplace=True)
            # Get statistics about one individual experiment
            expt_info.derived_stats = get_expt_stats(expt_df)

            expt_dc_list.append(DataContainer(df=expt_df, is_df=True))
        else:
            expt_dc_list.append(dc)

    size_in_megabytes = sys.getsizeof(expt_dc_list) / 10**6
    print(f"`expt_data` size in MB: {size_in_megabytes}")

    return expt_dc_list



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""The goal is to merge a collection of dataframes and number the common columns."""

import gc
import resource
import sys
from functools import reduce
from typing import Optional

import pandas as pd

from abvelocity.utils.check_df_validity import check_df_validity


def merge_dfs_number_cols(
    df_list: list[pd.DataFrame],
    on_cols: list[str],
    common_cols: Optional[list[str]] = None,
    how: str = "outer",
    add_tuple: bool = True,
    nan_replacement: Optional[str | float] = None,
    drop_numbered_cols: bool = False,
    incremental_df_delete: bool = True,
) -> pd.DataFrame:
    """A function to merge a list of pandas dataframes while numbering the common
        columns by numerics starting from 1.

        This is a different behaviour from Pandas which create {col}_x, {col}_y
        when merging two dataframes with a common column col.
        Here the purpose is to merge potentially more than two dataframes and end up
        with more predictable column names.

        This function will add
        {col}_1, {col}_2, {col}_3 etc as the new columns for each common column col.
        This function will also create tuple columns to combine all the values
        into one tuple as well if `add_tuple` is True.

        This function is cognizant of memory usage and takes steps to save
        memory space by

            - dropping used dataframes as it merges
            - dropping numbered columns as it is creating tuples (if desired).

    Args:
        df_list: A list of pandas dataframes to be merged.
        on_cols: The list of columns to join on.
        common_cols: A list of common columns to be numbered.
            If None, the assumption is all columns besides the `on_cols` are common.
        how: Specifies the method for merging.
            The default is "outer". It can also be "right" or "left".
        add_tuple: If tuple columns are to be added by combining the numbered columns.
        nan_replacement: If passed, it will replace NaN values with this value,
            only in `common_cols`.
        drop_numbered_cols: If it is desired to drop the numbered columns.
        incremental_df_delete: If it is desired to delete the dataframes as they are merged.
            Default is True.

    Returns:
        result: The merged dataframe with merged data for each common column, {col},  given in
            {col}_1, {col}_2, {col}_3, ...
            Also it could contain a tuple column with the same name {col} to store
            a tuple of all those values, if `add_tuple` is True.

    Alters:
        df_list: This argument will be altered to save memory as opposed to be copied.
            If you do not want to alter this argument, pass a copy.

    Raises:
        None.
    """
    all_cols = df_list[0].columns
    if common_cols is None:
        common_cols = [col for col in all_cols if col not in on_cols]

    df_num = len(df_list)

    # Data validation steps.
    err_trigger_source = "merge_dfs_number_cols"
    for i in range(df_num):
        df = df_list[i]
        # Checks the existence of `on_cols`.
        check_df_validity(
            df=df,
            needed_cols=on_cols,
            err_trigger_source=err_trigger_source,
            err_message=f"df_list[{i}] has missing columns for merge (`on_cols`).",
        )
        # Checks the existence of `common_cols`.
        check_df_validity(
            df=df,
            needed_cols=common_cols,
            err_trigger_source=err_trigger_source,
            err_message=f"df_list[{i}] has missing columns for merge (`common_cols`).",
        )

    # Below alters the name of the common columns
    # by adding a number: f"{col}_{i + 1}"" based on the order `i` in `df_list`
    for i in range(df_num):
        df = df_list[i]
        for col in common_cols:
            df_list[i].rename(columns={col: f"{col}_{i + 1}"}, inplace=True)

    def merge_two(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
        """Merges two dataframes.
        It deletes the input dataframes to save memory space.
        Args:
            df1 : First dataframe.
            df2 : Second dataframe.
        Returns:
            df: Merged dataframe.
        Raises:
            None.
        """
        df = pd.merge(df1, df2, on=on_cols, how=how)
        # Removes unnecessary data to retain memory space.
        if incremental_df_delete:
            del df1
            del df2
        # If `nan_replacement` is passed,
        # it will replace NaN values with this value.
        # Since we add numbers suffices to the common cols during merge,
        # we need to find the new column names, before we can replace NaNs.
        # The double for loop below is to achieve that goal.
        # Note that since the number of columns of df and common cols is small,
        # this will not be a computational problem.
        if nan_replacement is not None:
            for col in df.columns:
                for common_col in common_cols:
                    if col.startswith(common_col):
                        df.fillna({col: nan_replacement}, inplace=True)
        return df

    mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / (1024**3)  # Convert to GB
    print(f"\n*** Memory used in `merge_dfs_number_cols`, before merging dfs: {mem_usage:.4f} GB")

    df = reduce(merge_two, df_list)
    size_in_megabytes = sys.getsizeof(df) / 10**6
    print(f"\n***`df` size in MB: {size_in_megabytes}")

    mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / (1024**3)  # Convert to GB
    print(f"\n*** Memory used in `merge_dfs_number_cols`, after merging dfs: {mem_usage:.4f} GB")

    # Delete `df_list` and Garbage collection
    del df_list
    gc.collect()
    mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / (1024**3)  # Convert to GB
    print(
        f"\n*** Memory used in `merge_dfs_number_cols`, after del df_list and gc: {mem_usage:.4f} GB"
    )

    if add_tuple:
        for col in common_cols:
            added_cols = [f"{col}_{i + 1}" for i in range(df_num)]
            df[col] = list(zip(*[df[c] for c in added_cols]))
            # If the numbered columns are to be dropped,
            # it is done below as we are adding tuple columns to maintain memory space.
            # This will cause a few lines of (delibrate) repeated code below under `else`.
            if drop_numbered_cols:
                for col in added_cols:
                    del df[col]
    elif drop_numbered_cols:
        # This is the case where no tuple columns are added.
        # But we do want to drop the numbered columns.
        # This is a rare situation, but retains the generality of this function
        # as a generic merge function.
        for col in common_cols:
            added_cols = [f"{col}_{i + 1}" for i in range(df_num)]
            for col in added_cols:
                del df[col]

    return df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from dataclasses import dataclass, field
from typing import Callable, Optional

from abvelocity.param.metric import Metric


@dataclass
class MetricFamily:
    """This is a dataclass to store information on supported families of metrics.
    Such information is then used to get data.
    In order to add a new metric family developers need to:
        - (1) implement the get and process functions.
        - (2) add metric name to `~abvelocity.get_data.li_metric_families.METRIC_FAMILY_MAP`.
    """

    name: str
    """Metric family name."""
    get_metric_query: Optional[Callable] = None
    """Function to get a metric (SQL) query."""
    metrics: Optional[list[Metric]] = None
    """Default Metrics for the table."""
    get_metric_query_params: dict = field(default_factory=lambda: {})
    """Optional parameters to pass to `get_metric_query`."""
    process_expt_metric_df: Optional[Callable] = None
    """Optional function to post process the joined experiment and metric data."""
    process_expt_metric_df_params: dict = field(default_factory=lambda: {})
    """Optional parameters to pass to `process_expt_metric_df`."""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import os
from dataclasses import dataclass
from typing import Dict, List, Optional

import pandas as pd
import plotly.graph_objects as go

from abvelocity.journey.seq.count_seq_query import CountSeqQuery
from abvelocity.journey.viz.create_sunburst import create_sunburst
from abvelocity.param.io_param import IOParam


@dataclass
class SunburstResult:
    """Dataclass to hold the results of sunburst generation."""

    fig: Optional[go.Figure] = None
    df: Optional[pd.DataFrame] = None
    query: Optional[str] = None


class SunburstPlot:
    """Generates sunburst visualizations for sequence data."""

    def __init__(self, io_param: IOParam):
        """
        Initializes the SunburstPlot instance.

        Args:
            io_param (IOParam): Dataclass containing input/output parameters.
        """
        self.cursor = io_param.cursor
        self.print_to_html = io_param.print_to_html
        self.save_path = io_param.save_path
        self.file_name_suffix = io_param.file_name_suffix
        os.makedirs(self.save_path, exist_ok=True)

    def gen(
        self,
        table_name: str,
        value_col: str,
        conditions: List[str] = ["TRUE = TRUE"],
        max_seq_index: int = 5,
        count_distinct_col: Optional[str] = None,
        color_dict: Optional[Dict[str, str]] = None,
        additional_dimensions: Optional[List[str]] = None,
        top_n_dims: Optional[int] = 5,
    ) -> SunburstResult:
        """
        Generate a sunburst visualization for a given sequence table.
        It assumes the ordered events are provided in columns: s1, s2, ...,sk
        where k = `max_seq_index`.

        Args:
            table_name (str): Name of the joined table containing sequence data.
            value_col (str): Column name to be used for sunburst values.
            conditions (Optional[List[str]]): Optional list of SQL conditions.
            max_seq_index (int): Number of elements in the sequence / journey.
            count_distinct_col (Optional[str]): Column to count distinct values from.
            color_dict (Optional[Dict[str, str]]): A dictionary mapping labels to colors. If None, default colors are generated.
            additional_dimensions (Optional[List[str]]): Additional columns to group by in the sunburst plot. These columns will be
                                                     appended as extra levels in the sunburst hierarchy. Defaults to None.
            top_n_dims (Optional[int]): For each extra groupby column, only the top N most frequent values will be shown;
                                        all others will be grouped as 'Others'. This helps keep the plot readable. Defaults to 5.

        Returns:
            SunburstResult: Dataclass containing the generated chart data.
        """

        query = CountSeqQuery(
            table_name=table_name,
            max_seq_index=max_seq_index,
            count_distinct_col=count_distinct_col,
            extra_groupby_cols=additional_dimensions,
            conditions=conditions,
        ).gen()

        query += " ORDER BY percent DESC"
        df = self.cursor.get_df(query=query).df

        if df.empty:
            raise ValueError("df is empty, cannot plot.")

        # Consider the top N dimensions if specified and classifies the rest as 'Others' for the sake for readability.
        if top_n_dims and additional_dimensions:
            for dim in additional_dimensions:
                if dim in df.columns:
                    top_vals = df.groupby(dim)[value_col].sum().nlargest(top_n_dims).index
                    df[dim] = df[dim].where(df[dim].isin(top_vals), "Others")

        if additional_dimensions:
            for i, dim in enumerate(additional_dimensions, start=max_seq_index + 1):
                if dim not in df.columns:
                    continue
                # Prefix to keep labels unique and clear
                df[f"s{i}"] = df[dim].astype(str)
            # Bump the depth so create_sunburst sees them
            max_seq_index += len(additional_dimensions)

        df = df.fillna("end")
        print(f"\n\n\n*** df:\n\n\n {df.head()}")

        title = f"Sunburst plot: {table_name}: quantity:{value_col}: conditions: {conditions}"

        fig = create_sunburst(
            df=df,
            max_seq_index=max_seq_index,
            value_col=value_col,
            color_dict=color_dict,
            title=title,
        )

        file_name = os.path.join(
            self.save_path, f"sunburst{self.file_name_suffix}_{table_name.replace('.', '_')}.html"
        )
        if self.print_to_html:
            self.print_to_html(fig, file_name=file_name)

        return SunburstResult(fig=fig, df=df, query=query)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import pandas as pd
import plotly.graph_objects as go


def plot_2d_variants(
    df: pd.DataFrame,
    count_column: str = "variant_count",
    title: str = "2D Partition of Variants by Count",
) -> go.Figure:
    """
    Create an interactive bubble plot partitioning a 2D space based on two-dimensional categorical tuples using Plotly.

    Args:
        df: DataFrame with 'variant' (tuple) and count_column.
        count_column: Column name for bubble size (default: 'variant_count').
        title: Plot title.

    Returns:
        go.Figure: Plotly figure object.
    """
    # Extract unique labels for each dimension, ensuring 'nan' comes first
    dim1_labels = sorted(
        set(x[0] for x in df["variant"] if isinstance(x, tuple)), key=lambda x: (x != "nan", x)
    )
    dim2_labels = sorted(
        set(x[1] for x in df["variant"] if isinstance(x, tuple)), key=lambda x: (x != "nan", x)
    )

    # Map labels to numeric coordinates for plotting
    dim1_map = {label: idx for idx, label in enumerate(dim1_labels)}
    dim2_map = {label: idx for idx, label in enumerate(dim2_labels)}

    # Prepare plot data
    x_coords = [dim1_map[v[0]] for v in df["variant"]]
    y_coords = [dim2_map[v[1]] for v in df["variant"]]
    sizes = df[count_column].values
    labels = [f"({v[0]}, {v[1]})" for v in df["variant"]]

    # Scale sizes for better visualization (Plotly uses marker size, adjust as needed)
    max_count = max(sizes)
    sizes = [
        50 * (count / max_count) ** 0.5 for count in sizes
    ]  # Square root scaling for better visibility

    # Create bubble plot
    fig = go.Figure()

    fig.add_trace(
        go.Scatter(
            x=x_coords,
            y=y_coords,
            mode="markers",
            marker=dict(
                size=sizes,
                sizemode="diameter",
                sizeref=max(sizes) / 50,  # Adjust max bubble size
                color=sizes,
                colorscale="Viridis",
                showscale=True,
                colorbar=dict(title=count_column),
            ),
            text=labels,
            customdata=df[count_column].values,
            hovertemplate="Variant: %{text}<br>Count: %{customdata:,.0f}<extra></extra>",
        )
    )

    # Update layout
    fig.update_layout(
        title=title,
        xaxis=dict(
            title="Dimension 1",
            tickvals=list(range(len(dim1_labels))),
            ticktext=dim1_labels,
            range=[-0.5, len(dim1_labels) - 0.5],
        ),
        yaxis=dict(
            title="Dimension 2",
            tickvals=list(range(len(dim2_labels))),
            ticktext=dim2_labels,
            range=[-0.5, len(dim2_labels) - 0.5],
        ),
        showlegend=False,
        width=800,
        height=600,
    )

    return fig



# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# author: Reza Hosseini
from typing import Optional

import pandas as pd


def df_to_markdown(
    df: pd.DataFrame,
    heading: Optional[str] = None,
    top_paragraphs: list[str] = [],
    file_name: Optional[str] = None,
    caption: str = "",
) -> str:
    """Converts a pandas dataframe to a markdown string.
        This is useful to generate quick readable markdown reports.

    Args:
        df: A pandas dataframe.
        heading: A string for the heading of the markdown.
            The default is None, in which case no heading will be included.
        top_paragraphs: A list of strings for the top paragraphs of the markdown.
            The default is an empty list.
        file_name: A string for the name of the markdown file.
            If None, no file will be written.
        caption: A string for the caption of the table.
            The default is an empty string.

    Returns:
        A string of the markdown.
    """
    # Create the markdown table representation
    df_md = df.to_markdown(index=False)

    # Adding the caption if provided
    if caption:
        df_md = f"**{caption}**\n\n" + df_md

    md_str = ""
    if heading is not None:
        md_str += f"# {heading}\n\n"

    for paragraph in top_paragraphs:
        md_str += f"{paragraph}\n\n"

    md_str += df_md + "\n"

    if file_name is not None:
        with open(file_name, "w") as f:
            f.write(md_str)
            print(f"{file_name} was created.")

    return md_str



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from dataclasses import dataclass
from typing import Optional

import numpy as np

from abvelocity.param.constants import CATEG_NAN_VALUE


@dataclass
class Variant:
    value: str | tuple[str]
    """The variant label/value in the data which can be a:

        - `str` (for simple experiments) or
        - a `tuple[str]` (for multi-experiments).

    """
    name: Optional[str] = None
    """The name of the variant."""

    def __post_init__(self):
        if self.name is None:
            if isinstance(self.value, str):
                self.name = self.value
            else:
                self.name = "(" + ", ".join(self.value) + ")"

    def __eq__(self, other):
        """We only require the values being equal."""
        if not isinstance(other, Variant):
            return NotImplemented
        # Compare only the value
        return self.value == other.value


@dataclass
class TriggerState:
    value: Optional[bool | tuple[bool]] = None
    """The trigger state of the variant in the data where its `value` can be a:

            - `bool` (for simple experiments) or
            - a `tuple[bool]` (for multi-experiments).

    This determines which experiment is triggered for that unit of the experiment data.

    It also includes an `overall_state` which is a boolean value indicating whether any of the components of the tuple is `True`.
    This is useful for multi-experiments where we want to know if any of the experiments are triggered.
    Note that `overall_value` is inferred from the value if not passed explicitly.
    """
    overall_value: Optional[bool] = None
    """
    bool indicating whether any of the components of the tuple is `True`.
    This is useful for multi-experiments where we want to know if any of the experiments are triggered.
    """
    name: Optional[str] = None
    """The name of the trigger state."""

    def __post_init__(self):
        if self.name is None:
            if isinstance(self.value, bool) or isinstance(self.value, np.bool_):
                self.name = str(self.value)
            else:
                self.name = "(" + ", ".join([str(trigger) for trigger in self.value]) + ")"

        if self.value is not None and self.overall_value is None:
            if isinstance(self.value, bool) or isinstance(self.value, np.bool_):
                self.overall_value = self.value
            else:
                self.overall_value = any(self.value)


@dataclass
class VariantList:
    variants: list[Variant]
    """The list of variants in the list."""
    name: Optional[str] = None
    """The name of the variant list."""

    def __post_init__(self):
        if len(self.variants) == 0:
            raise ValueError("`.variants` in `VariantList` field cannot be an empty list.")
        first_type = type(self.variants[0].value)
        tuple_length = (
            len(self.variants[0].value) if isinstance(self.variants[0].value, tuple) else None
        )

        for variant in self.variants:
            if type(variant.value) != first_type:
                raise ValueError(
                    "All elements in the VariantList must be of the same type (either all str or all tuple)."
                )
            if isinstance(variant.value, tuple):
                if len(variant.value) != tuple_length:
                    raise ValueError(
                        "All tuple elements in the VariantList must be of the same size."
                    )

        if self.name is None:
            self.name = "[" + ", ".join([variant.name for variant in self.variants]) + "]"


@dataclass
class ComparisonPair:
    treatment: VariantList
    """The list of variants in treatment.
    Note that the treatment might be a collection of variants."""
    control: VariantList
    """The list of variants in the control.
    Note that the control might be a collection of variants."""
    name: Optional[str] = None
    """The name of the comparison pair."""

    def remove_common_variants(self):
        """
        - Removes common variants from both arms.
        - Throws an error if either or both of arms have zero variants.
            This could happen after removing common variants (which is also user input error)."""

        # Extract the variants from both arms
        treatment_variant_values = [Variant.value for Variant in self.treatment.variants]
        control_variant_values = [Variant.value for Variant in self.control.variants]

        # Error handling for having non-empty variants on both arms, after removal.
        # If either or both arms have empty variants, we raise error.
        # This implies one is subset of the other.
        set1, set2 = set(treatment_variant_values), set(control_variant_values)
        diff1 = set1 - set2  # Elements in the first list but not in the second
        diff2 = set2 - set1  # Elements in the first list but not in the second
        if not diff1 or not diff2:
            raise ValueError(
                "At least one of the arms does not have any variants."
                " This might happen due to removal of common variants."
                " Which implies one of the arms variants are subset of the other (or both are the same)."
                " These are the original variants passed by user:"
                f" treatment: {treatment_variant_values}"
                f" control: {control_variant_values}"
            )

        # `.name` is not updated by design to be faithful to previous name
        # This does have the caveat that if the name lists old variants adn we drop some,
        # it can create some confusion.
        # But it might be better to keep it, as the comparison still is a comparison
        # for the original lists of variants and the drop is only done to increase statistical power
        self.treatment = VariantList(
            variants=[
                variant
                for variant in self.treatment.variants
                if variant.value not in control_variant_values
            ],
            name=self.treatment.name,
        )

        self.control = VariantList(
            variants=[
                variant
                for variant in self.control.variants
                if variant.value not in treatment_variant_values
            ],
            name=self.control.name,
        )

    def __post_init__(self):
        """
        This accomplishes:
            - Throws an error if either or both of arms have zero variants.
                This could happen after removing common variants.
            - Removes common variants.
            - Assigns a name based on the input, if it's not passed.
        """
        # Remove the common variants.
        # Does error handling.
        self.remove_common_variants()

        combined_variants = self.treatment.variants + self.control.variants

        first_type = type(combined_variants[0].value)
        tuple_length = (
            len(combined_variants[0].value)
            if isinstance(combined_variants[0].value, tuple)
            else None
        )

        for variant in combined_variants:
            if type(variant.value) != first_type:
                raise ValueError(
                    "All elements in treatment and control must be of the same type (either all str or all tuple)."
                )
            if isinstance(variant.value, tuple):
                if len(variant.value) != tuple_length:
                    raise ValueError(
                        "All tuple elements in treatment and control must be of the same size."
                    )

        if self.name is None:
            self.name = f"{self.treatment.name} versus {self.control.name}"


def variant_to_trigger_state(variant: Variant) -> TriggerState:
    """Converts a given variant to a trigger state.
    Example 1: (single experiment)
        - variant.value: "v1"
        - trigger_state.value: True

    Example 2: (multi-experiment)
        - variant.value: ("v1", "w1", "x1")
        - trigger_state.value: (True, True, True)

    Example 3: (multi-experiment)
        - variant.value: ("v1", "w1", "nan")
        - trigger_state.value: (True, True, False)

    Args:
        variant: The variant to convert.

    Returns:
        The trigger state.
    """

    if isinstance(variant.value, str):
        if variant.value == CATEG_NAN_VALUE:
            return TriggerState(value=False)
        return TriggerState(value=True)

    trigger_state_value = tuple([True if v != CATEG_NAN_VALUE else False for v in variant.value])

    return TriggerState(value=trigger_state_value)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini


import sys
import time
from dataclasses import dataclass
from typing import Any, Optional

import pandas as pd


@dataclass
class SqlQueryResult:
    """This is a dataclass to store a SQL query dataframe, time taken and the size."""

    df: Optional[pd.DataFrame] = None
    """A `pandas` dataframe which includes the data returned by the SQL query or None."""
    query_time_taken: Optional[float] = 0
    """The time taken to run the query in seconds (float)."""
    size_in_megabytes: Optional[float] = 0
    """The memory size of resulting dataframe."""


class Cursor:
    """
    A generic cursor class that provides a consistent interface for executing queries
    and fetching results, adhering to the PEP 249 standard where possible.
    """

    def __init__(self, data_source: Any):
        """
        Initializes the cursor with a data source.
        """
        self.data_source = data_source
        self.description = None
        self.results = None

    def execute(self, query: str):
        """
        Executes a query on the data source.
        Needs to update
            self.description
            self.results
        """
        pass

    def fetchall(self):
        """
        Fetches all rows of a query result as a list of tuples or lists.
        """
        return []

    def fetchone(self):
        """
        Fetches the next row of a query result set.
        """
        return None

    def get_df(self, query: str) -> SqlQueryResult:
        """
        Executes a query and returns the results as a SqlQueryResult dataclass instance.
        """
        start_time = time.time()
        self.execute(query)
        if self.description:
            columns = [col_info[0] for col_info in self.description]
            print(f"\n*** col_names: {columns}")

            print("\n*** create dataframe")
            df = pd.DataFrame(self.results, columns=columns)
            end_time = time.time()
            query_time_taken = end_time - start_time
            size_in_megabytes = sys.getsizeof(df) / 10**6 if df is not None else 0

            return SqlQueryResult(
                df=df, query_time_taken=query_time_taken, size_in_megabytes=size_in_megabytes
            )

        return SqlQueryResult(df=None, query_time_taken=0, size_in_megabytes=0)

    def execute_multi(self, query: str) -> None:
        """
        Executes multiple queries separated by ';'.
        """
        sub_queries = query.strip().split(";")

        for sub_query in sub_queries:
            sub_query = sub_query.strip()
            if sub_query:  # Ensure it's not empty
                print(f"\n***\n sub_query: {sub_query}")
                self.execute(sub_query)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
import re


def assert_query_is_equal(query1: str, query2: str) -> bool:
    """Compares two queries, ignoring whitespace, newlines, and case for keywords,
    but preserving case for string literals.
    """

    def standardize_query(query: str) -> str:
        # 1. Remove comments
        query = re.sub(r"--.*$", "", query, flags=re.MULTILINE)
        query = re.sub(r"/\*.*?\*/", "", query, flags=re.DOTALL)

        # 2. Replace all newlines with spaces
        query = query.replace("\n", " ")

        # 3. Normalize whitespace and convert to single spaces
        query = re.sub(r"\s+", " ", query).strip()

        # 4. Convert keywords to uppercase (after whitespace normalization)
        def replace_keyword(match):
            keyword = match.group(0)
            if keyword.startswith("'") or keyword.startswith('"'):
                return keyword  # String literal, keep as is
            return keyword.upper()

        query = re.sub(r"('[^']*'|\"[^\"]*\"|[a-zA-Z_]+)", replace_keyword, query)
        return query

    standardized_query1 = standardize_query(query1)
    standardized_query2 = standardize_query(query2)

    if standardized_query1 != standardized_query2:
        raise AssertionError(
            f"\n***\nquery1: {standardized_query1}\n is not same as \nquery2: {standardized_query2}"
        )

    return True



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from typing import Optional

import numpy as np
import pandas as pd


def round_float_cols(df: pd.DataFrame, rounding_digits: int = 4, cols: Optional[list[str]] = None):
    """Round the float columns of the dataframe based on `rounding_digits`.

    Args:
        df: Dataframe to round the columns of.
        rounding_digits: Number of digits to round to.
            Default is 4.
        cols: Columns to round.
            Default is None, which means all columns.

    Returns:
        None

    Alters:
        df: The dataframe passed in.

    """
    if cols is None:
        cols = list(df.columns)

    for col in cols:
        if col in df.columns and df[col].dtype == "float64":
            df[col] = df[col].apply(lambda x: round(x, rounding_digits))

    return None


def round_tuple_cols(df: pd.DataFrame, rounding_digits: int = 4, cols: Optional[list[str]] = None):
    """Round the columns that are tuples in the dataframe.
        In this case, we do two checks to pick the revelant columns:

            - If the column is `CI_COL` or `CI_PERCENT_COL`
            - the column consist of numpy array as elements.

    Args:
        df: Dataframe to round the columns of.

    Returns:
        None

    Alters:
        df: The dataframe passed in.

    """
    if cols is None:
        cols = list(df.columns)

    for col in cols:
        if col in df.columns:
            # Check if the column is a tuple of floats.
            # This could be either a numpy array or tuple of floats.
            # We include both possibilities.
            if (
                isinstance(df[col].values[0], np.ndarray)
                or isinstance(df[col].values[0], tuple)
                and isinstance(df[col].values[0][0], float)
            ):
                df[col] = df[col].apply(lambda x: tuple([round(y, rounding_digits) for y in x]))

    return None


def round_df(
    df,
    rounding_digits: int = 4,
    float_cols: Optional[list[str]] = None,
    tuple_cols: Optional[list[str]] = None,
):
    """Round the float columns and tuple columns of the dataframe.

    Args:
        df: Dataframe to round the columns of.
        rounding_digits: Number of digits to round to.
            Default is 4.
        float_cols: Columns to round that are floats.
            Default is None, which means all columns.
        tuple_cols: Columns to round that are tuples.
            Default is None, which means all columns.

    Returns:
        None

    Alters:
        df: The dataframe passed in.

    """
    round_float_cols(df=df, rounding_digits=rounding_digits, cols=float_cols)
    round_tuple_cols(df=df, rounding_digits=rounding_digits, cols=tuple_cols)
